from typing import List, Tuple, Union

import numpy as np

TOLERANCE = 1e-8


class GeometryPolygon:
    def __init__(
        self,
        points: Union[np.ndarray, List[Tuple[float, float]]],
        is_rectangle: bool = False,
    ):
        self.points = (
            points if isinstance(points, np.ndarray) else np.array(points)
        )
        self.is_rectangle = is_rectangle
        points_x = self.points[:, 0]
        points_y = self.points[:, 1]
        if is_rectangle:
            self.signed_area = np.abs(self.points[2] - self.points[0]).prod()
            self.area = self.signed_area
        else:
            self.signed_area = (
                points_x @ np.roll(points_y, -1)
                - points_x @ np.roll(points_y, 1)
            ) / 2
            self.area = np.abs(self.signed_area)

    def __len__(self):
        return len(self.points)

    def __getitem__(self, idx):
        return self.points[idx]

    def __repr__(self) -> str:
        return f"GeometryPolygon({self.points})"


# alpha * a1 + (1 - alpha) * a2 = beta * b1 + (1 - beta) * b2
def segment_intersection(
    segment1: Tuple[np.ndarray, np.ndarray],
    segment2: Tuple[np.ndarray, np.ndarray],
) -> Tuple[float, float, np.ndarray]:
    a1, a2 = segment1
    b1, b2 = segment2
    x2_x2 = b2[0] - a2[0]
    y2_y2 = b2[1] - a2[1]
    x1x2 = a1[0] - a2[0]
    y1y2 = a1[1] - a2[1]
    y1_y2_ = b1[1] - b2[1]
    x1_x2_ = b1[0] - b2[0]

    if np.abs(y1_y2_ * x1x2 - x1_x2_ * y1y2) < TOLERANCE:
        beta = 1.0
    else:
        beta = (x2_x2 * y1y2 - y2_y2 * x1x2) / (y1_y2_ * x1x2 - x1_x2_ * y1y2)

    if x1x2 == 0:
        alpha = (y2_y2 + y1_y2_ * beta) / (y1y2 + TOLERANCE)
    else:
        alpha = (x2_x2 + x1_x2_ * beta) / (x1x2 + TOLERANCE)

    return alpha, beta, alpha * a1 + (1 - alpha) * a2


def convex_polygon_intersection_area(
    polygon_a: GeometryPolygon, polygon_b: GeometryPolygon
) -> float:
    # pylint: disable=R0912
    sa = polygon_a.signed_area
    sb = polygon_b.signed_area
    if sa * sb < 0:
        sign = -1
    else:
        sign = 1
    na = len(polygon_a)
    nb = len(polygon_b)
    ps = []  # point set
    for i in range(na):
        a1 = polygon_a[i - 1]
        a2 = polygon_a[i]
        flag = False
        sum_s = 0
        for j in range(nb):
            b1 = polygon_b[j - 1]
            b2 = polygon_b[j]
            sum_s += np.abs(GeometryPolygon([a1, b1, b2]).signed_area)

        if np.abs(np.abs(sum_s) - np.abs(sb)) < TOLERANCE:
            flag = True

        if flag:
            ps.append(a1)
        for j in range(nb):
            b1 = polygon_b[j - 1]
            b2 = polygon_b[j]
            a, b, p = segment_intersection((a1, a2), (b1, b2))
            if 0 < a < 1 and 0 < b < 1:
                ps.append(p)

    for i in range(nb):
        a1 = polygon_b[i - 1]
        a2 = polygon_b[i]
        flag = False
        sum_s = 0
        for j in range(na):
            b1 = polygon_a[j - 1]
            b2 = polygon_a[j]
            sum_s += np.abs(GeometryPolygon([a1, b1, b2]).signed_area)
        if np.abs(np.abs(sum_s) - np.abs(sa)) < TOLERANCE:
            flag = True
        if flag:
            ps.append(a1)

    def unique(ar):
        res = []
        for i, a in enumerate(ar):
            if np.abs(a - ar[i - 1]).sum() > TOLERANCE:
                res.append(a)

        return res

    ps = sorted(ps, key=lambda x: (x[0] + TOLERANCE * x[1]))
    ps = unique(ps)

    if len(ps) == 0:
        return 0

    tmp = ps[0]

    res = []
    res.append(tmp)
    ps = sorted(
        ps[1:],
        key=lambda x: -((x - tmp) @ np.array((0, 1)) / len(x - tmp)),
    )
    res.extend(ps)

    return GeometryPolygon(res).signed_area * sign


def area(box):
    if box[2] <= box[0] or box[3] <= box[1]:
        return 0
    return (box[2] - box[0]) * (box[3] - box[1])


def iou(box_a, box_b):
    box_c = intersection(box_a, box_b)
    return area(box_c) / (area(box_a) + area(box_b) - area(box_c))


def intersection(box_a, box_b):
    """boxes are left, top, right, bottom where left < right and top < bottom"""
    box_c = [
        max(box_a[0], box_b[0]),
        max(box_a[1], box_b[1]),
        min(box_a[2], box_b[2]),
        min(box_a[3], box_b[3]),
    ]
    return box_c


def rectangle_intersection_area(
    polygon_a: GeometryPolygon, polygon_b: GeometryPolygon
) -> float:
    minx_a, miny_a = np.min(polygon_a.points, axis=0)
    maxx_a, maxy_a = np.max(polygon_a.points, axis=0)
    minx_b, miny_b = np.min(polygon_b.points, axis=0)
    maxx_b, maxy_b = np.max(polygon_b.points, axis=0)

    minx_c = max(minx_a, minx_b)
    miny_c = max(miny_a, miny_b)
    maxx_c = min(maxx_a, maxx_b)
    maxy_c = min(maxy_a, maxy_b)
    return max(maxx_c - minx_c, 0) * max(maxy_c - miny_c, 0)


def polygon_intersection_area(
    polygon_a: GeometryPolygon, polygon_b: GeometryPolygon
) -> float:
    if polygon_a.is_rectangle and polygon_b.is_rectangle:
        return rectangle_intersection_area(polygon_a, polygon_b)

    na = len(polygon_a)
    nb = len(polygon_b)
    res = 0.0
    for i in range(1, na - 1):
        sa = polygon_a[[0, i, i + 1]]
        for j in range(1, nb - 1):
            sb = polygon_b[[0, j, j + 1]]
            tmp = convex_polygon_intersection_area(
                GeometryPolygon(sa), GeometryPolygon(sb)
            )
            res += tmp

    return np.abs(res)



import pkg_resources

nucleus_client_version = pkg_resources.get_distribution(
    "scale-nucleus"
).version

INFRA_FLAKE_MESSAGES = [
    "downstream duration timeout",
    "upstream connect error or disconnect/reset before headers. reset reason: local reset",
]


class ModelCreationError(Exception):
    def __init__(self, message="Could not create the model"):
        self.message = message
        super().__init__(self.message)


class ModelRunCreationError(Exception):
    def __init__(self, message="Could not create the model run"):
        self.message = message
        super().__init__(self.message)


class NotFoundError(Exception):
    def __init__(
        self, message="Could not open file. Check the path or if it exists."
    ):
        self.message = message
        super().__init__(self.message)


class DatasetItemRetrievalError(Exception):
    def __init__(self, message="Could not retrieve dataset items"):
        self.message = message
        super().__init__(self.message)


class NucleusAPIError(Exception):
    def __init__(
        self, endpoint, command, requests_response=None, aiohttp_response=None
    ):
        self.message = f"Your client is on version {nucleus_client_version}. If you have not recently done so, please make sure you have updated to the latest version of the client by running pip install --upgrade scale-nucleus\n"
        if requests_response is not None:
            self.status_code = requests_response.status_code
            self.message += f"Tried to {command.__name__} {endpoint}, but received {requests_response.status_code}: {requests_response.reason}."
            if hasattr(requests_response, "text"):
                if requests_response.text:
                    self.message += (
                        f"\nThe detailed error is:\n{requests_response.text}"
                    )

        if aiohttp_response is not None:
            status, reason, data = aiohttp_response
            self.status_code = status
            self.message += f"Tried to {command.__name__} {endpoint}, but received {status}: {reason}."
            if data:
                self.message += f"\nThe detailed error is:\n{data}"

        if any(
            infra_flake_message in self.message
            for infra_flake_message in INFRA_FLAKE_MESSAGES
        ):
            self.message += "\n This likely indicates temporary downtime of the API, please try again in a minute or two"
        super().__init__(self.message)


class NoAPIKey(Exception):
    def __init__(
        self,
        message="You need to pass an API key to the NucleusClient or set the environment variable NUCLEUS_API_KEY",
    ):
        self.message = message
        super().__init__(self.message)



from enum import Enum
from typing import TYPE_CHECKING, Dict

if TYPE_CHECKING:
    from . import NucleusClient


# Wording set to match with backend enum
class ExportMetadataType(Enum):
    SCENES = "scene"
    DATASET_ITEMS = "item"


class MetadataManager:
    """
    Helper class for managing metadata updates on a scene or dataset item.
    Do not call directly, use the dataset class methods: `update_scene_metadata` or `update_item_metadata`
    """

    def __init__(
        self,
        dataset_id: str,
        client: "NucleusClient",
        raw_mappings: Dict[str, dict],
        level: ExportMetadataType,
    ):
        self.dataset_id = dataset_id
        self._client = client
        self.raw_mappings = raw_mappings
        self.level = level

        self._payload = self._format_mappings()

    def _format_mappings(self):
        payload = []
        for ref_id, meta in self.raw_mappings.items():
            payload.append({"reference_id": ref_id, "metadata": meta})
        return payload

    def update(self):
        payload = {"metadata": self._payload, "level": self.level.value}
        resp = self._client.make_request(
            payload=payload, route=f"dataset/{self.dataset_id}/metadata"
        )
        return resp



import functools
import os

import nucleus


@functools.lru_cache()
def init_client():
    api_key = os.environ.get("NUCLEUS_API_KEY", None)
    if api_key:
        client = nucleus.NucleusClient(api_key)
    else:
        raise RuntimeError("No NUCLEUS_API_KEY set")
    return client



from typing import List, Optional

from nucleus.pydantic_base import DictCompatibleImmutableModel


class ScenesListEntry(DictCompatibleImmutableModel):
    """/dataset/{dataset_id}/scenes_list nested payload"""

    id: str
    reference_id: str
    type: str
    metadata: Optional[dict] = {}


class ScenesList(DictCompatibleImmutableModel):
    """/dataset/{dataset_id}/scenes_list payload"""

    scenes: List[ScenesListEntry]



import sys
from abc import abstractmethod
from typing import List, Optional, Union

import numpy as np

from nucleus.annotation import AnnotationList, BoxAnnotation, PolygonAnnotation
from nucleus.prediction import BoxPrediction, PolygonPrediction, PredictionList

from .base import Metric, ScalarResult
from .filtering import ListOfAndFilters, ListOfOrAndFilters
from .filters import confidence_filter, polygon_label_filter
from .metric_utils import compute_average_precision
from .polygon_utils import (
    BoxOrPolygonAnnotation,
    BoxOrPolygonPrediction,
    get_true_false_positives_confidences,
    group_boxes_or_polygons_by_label,
    iou_assignments,
    label_match_wrapper,
    num_true_positives,
)


class PolygonMetric(Metric):
    """Abstract class for metrics of box and polygons.

    The PolygonMetric class automatically filters incoming annotations and
    predictions for only box and polygon annotations. It also filters
    predictions whose confidence is less than the provided confidence_threshold.
    Finally, it provides support for enforcing matching labels. If
    `enforce_label_match` is set to True, then annotations and predictions will
    only be matched if they have the same label.

    To create a new concrete PolygonMetric, override the `eval` function
    with logic to define a metric between box/polygon annotations and predictions.
    ::

        from typing import List
        from nucleus import BoxAnnotation, Point, PolygonPrediction
        from nucleus.annotation import AnnotationList
        from nucleus.prediction import PredictionList
        from nucleus.metrics import ScalarResult, PolygonMetric
        from nucleus.metrics.polygon_utils import BoxOrPolygonAnnotation, BoxOrPolygonPrediction

        class MyPolygonMetric(PolygonMetric):
            def eval(
                self,
                annotations: List[BoxOrPolygonAnnotation],
                predictions: List[BoxOrPolygonPrediction],
            ) -> ScalarResult:
                value = (len(annotations) - len(predictions)) ** 2
                weight = len(annotations)
                return ScalarResult(value, weight)

        box_anno = BoxAnnotation(
            label="car",
            x=0,
            y=0,
            width=10,
            height=10,
            reference_id="image_1",
            annotation_id="image_1_car_box_1",
            metadata={"vehicle_color": "red"}
        )

        polygon_pred = PolygonPrediction(
            label="bus",
            vertices=[Point(100, 100), Point(150, 200), Point(200, 100)],
            reference_id="image_2",
            annotation_id="image_2_bus_polygon_1",
            confidence=0.8,
            metadata={"vehicle_color": "yellow"}
        )

        annotations = AnnotationList(box_annotations=[box_anno])
        predictions = PredictionList(polygon_predictions=[polygon_pred])
        metric = MyPolygonMetric()
        metric(annotations, predictions)
    """

    def __init__(
        self,
        enforce_label_match: bool = False,
        confidence_threshold: float = 0.0,
        annotation_filters: Optional[
            Union[ListOfOrAndFilters, ListOfAndFilters]
        ] = None,
        prediction_filters: Optional[
            Union[ListOfOrAndFilters, ListOfAndFilters]
        ] = None,
    ):
        """Initializes PolygonMetric abstract object.

        Args:
            enforce_label_match: whether to enforce that annotation and prediction labels must match. Default False
            confidence_threshold: minimum confidence threshold for predictions. Must be in [0, 1]. Default 0.0
            annotation_filters: Filter predicates. Allowed formats are:
                ListOfAndFilters where each Filter forms a chain of AND predicates.
                    or
                ListOfOrAndFilters where Filters are expressed in disjunctive normal form (DNF), like
                [[MetadataFilter("short_haired", "==", True), FieldFilter("label", "in", ["cat", "dog"]), ...].
                DNF allows arbitrary boolean logical combinations of single field predicates. The innermost structures
                each describe a single column predicate. The list of inner predicates is interpreted as a conjunction
                (AND), forming a more selective `and` multiple field predicate.
                Finally, the most outer list combines these filters as a disjunction (OR).
            prediction_filters: Filter predicates. Allowed formats are:
                ListOfAndFilters where each Filter forms a chain of AND predicates.
                    or
                ListOfOrAndFilters where Filters are expressed in disjunctive normal form (DNF), like
                [[MetadataFilter("short_haired", "==", True), FieldFilter("label", "in", ["cat", "dog"]), ...].
                DNF allows arbitrary boolean logical combinations of single field predicates. The innermost structures
                each describe a single column predicate. The list of inner predicates is interpreted as a conjunction
                (AND), forming a more selective `and` multiple field predicate.
                Finally, the most outer list combines these filters as a disjunction (OR).
        """
        super().__init__(annotation_filters, prediction_filters)
        self.enforce_label_match = enforce_label_match
        assert 0 <= confidence_threshold <= 1
        self.confidence_threshold = confidence_threshold

    @abstractmethod
    def eval(
        self,
        annotations: List[BoxOrPolygonAnnotation],
        predictions: List[BoxOrPolygonPrediction],
    ) -> ScalarResult:
        # Main evaluation function that subclasses must override.
        pass

    def aggregate_score(self, results: List[ScalarResult]) -> ScalarResult:  # type: ignore[override]
        return ScalarResult.aggregate(results)

    def call_metric(
        self, annotations: AnnotationList, predictions: PredictionList
    ) -> ScalarResult:
        if self.confidence_threshold > 0:
            predictions = confidence_filter(
                predictions, self.confidence_threshold
            )
        polygon_annotations: List[Union[BoxAnnotation, PolygonAnnotation]] = []
        polygon_annotations.extend(annotations.box_annotations)
        polygon_annotations.extend(annotations.polygon_annotations)
        polygon_predictions: List[Union[BoxPrediction, PolygonPrediction]] = []
        polygon_predictions.extend(predictions.box_predictions)
        polygon_predictions.extend(predictions.polygon_predictions)

        eval_fn = label_match_wrapper(self.eval)
        result = eval_fn(
            polygon_annotations,
            polygon_predictions,
            enforce_label_match=self.enforce_label_match,
        )
        return result


class PolygonIOU(PolygonMetric):
    """Calculates the average IOU between box or polygon annotations and predictions.
    ::

        from nucleus import BoxAnnotation, Point, PolygonPrediction
        from nucleus.annotation import AnnotationList
        from nucleus.prediction import PredictionList
        from nucleus.metrics import PolygonIOU

        box_anno = BoxAnnotation(
            label="car",
            x=0,
            y=0,
            width=10,
            height=10,
            reference_id="image_1",
            annotation_id="image_1_car_box_1",
            metadata={"vehicle_color": "red"}
        )

        polygon_pred = PolygonPrediction(
            label="bus",
            vertices=[Point(100, 100), Point(150, 200), Point(200, 100)],
            reference_id="image_2",
            annotation_id="image_2_bus_polygon_1",
            confidence=0.8,
            metadata={"vehicle_color": "yellow"}
        )

        annotations = AnnotationList(box_annotations=[box_anno])
        predictions = PredictionList(polygon_predictions=[polygon_pred])
        metric = PolygonIOU()
        metric(annotations, predictions)
    """

    # TODO: Remove defaults once these are surfaced more cleanly to users.
    def __init__(
        self,
        enforce_label_match: bool = False,
        iou_threshold: float = 0.0,
        confidence_threshold: float = 0.0,
        annotation_filters: Optional[
            Union[ListOfOrAndFilters, ListOfAndFilters]
        ] = None,
        prediction_filters: Optional[
            Union[ListOfOrAndFilters, ListOfAndFilters]
        ] = None,
    ):
        """Initializes PolygonIOU object.

        Args:
            enforce_label_match: whether to enforce that annotation and prediction labels must match. Defaults to False
            iou_threshold: IOU threshold to consider detection as valid. Must be in [0, 1]. Default 0.0
            confidence_threshold: minimum confidence threshold for predictions. Must be in [0, 1]. Default 0.0
            annotation_filters: Filter predicates. Allowed formats are:
                ListOfAndFilters where each Filter forms a chain of AND predicates.
                    or
                ListOfOrAndFilters where Filters are expressed in disjunctive normal form (DNF), like
                [[MetadataFilter("short_haired", "==", True), FieldFilter("label", "in", ["cat", "dog"]), ...].
                DNF allows arbitrary boolean logical combinations of single field predicates. The innermost structures
                each describe a single column predicate. The list of inner predicates is interpreted as a conjunction
                (AND), forming a more selective `and` multiple field predicate.
                Finally, the most outer list combines these filters as a disjunction (OR).
            prediction_filters: Filter predicates. Allowed formats are:
                ListOfAndFilters where each Filter forms a chain of AND predicates.
                    or
                ListOfOrAndFilters where Filters are expressed in disjunctive normal form (DNF), like
                [[MetadataFilter("short_haired", "==", True), FieldFilter("label", "in", ["cat", "dog"]), ...].
                DNF allows arbitrary boolean logical combinations of single field predicates. The innermost structures
                each describe a single column predicate. The list of inner predicates is interpreted as a conjunction
                (AND), forming a more selective `and` multiple field predicate.
                Finally, the most outer list combines these filters as a disjunction (OR).
        """
        assert (
            0 <= iou_threshold <= 1
        ), "IoU threshold must be between 0 and 1."
        self.iou_threshold = iou_threshold
        super().__init__(
            enforce_label_match,
            confidence_threshold,
            annotation_filters,
            prediction_filters,
        )

    def eval(
        self,
        annotations: List[BoxOrPolygonAnnotation],
        predictions: List[BoxOrPolygonPrediction],
    ) -> ScalarResult:
        iou_assigns = iou_assignments(
            annotations, predictions, self.iou_threshold
        )
        weight = max(len(annotations), len(predictions))
        avg_iou = iou_assigns.sum() / max(weight, sys.float_info.epsilon)
        return ScalarResult(avg_iou, weight)


class PolygonPrecision(PolygonMetric):
    """Calculates the precision between box or polygon annotations and predictions.
    ::

        from nucleus import BoxAnnotation, Point, PolygonPrediction
        from nucleus.annotation import AnnotationList
        from nucleus.prediction import PredictionList
        from nucleus.metrics import PolygonPrecision

        box_anno = BoxAnnotation(
            label="car",
            x=0,
            y=0,
            width=10,
            height=10,
            reference_id="image_1",
            annotation_id="image_1_car_box_1",
            metadata={"vehicle_color": "red"}
        )

        polygon_pred = PolygonPrediction(
            label="bus",
            vertices=[Point(100, 100), Point(150, 200), Point(200, 100)],
            reference_id="image_2",
            annotation_id="image_2_bus_polygon_1",
            confidence=0.8,
            metadata={"vehicle_color": "yellow"}
        )

        annotations = AnnotationList(box_annotations=[box_anno])
        predictions = PredictionList(polygon_predictions=[polygon_pred])
        metric = PolygonPrecision()
        metric(annotations, predictions)
    """

    # TODO: Remove defaults once these are surfaced more cleanly to users.
    def __init__(
        self,
        enforce_label_match: bool = False,
        iou_threshold: float = 0.5,
        confidence_threshold: float = 0.0,
        annotation_filters: Optional[
            Union[ListOfOrAndFilters, ListOfAndFilters]
        ] = None,
        prediction_filters: Optional[
            Union[ListOfOrAndFilters, ListOfAndFilters]
        ] = None,
    ):
        """Initializes PolygonPrecision object.

        Args:
            enforce_label_match: whether to enforce that annotation and prediction labels must match. Defaults to False
            iou_threshold: IOU threshold to consider detection as valid. Must be in [0, 1]. Default 0.5
            confidence_threshold: minimum confidence threshold for predictions. Must be in [0, 1]. Default 0.0
            annotation_filters: Filter predicates. Allowed formats are:
                ListOfAndFilters where each Filter forms a chain of AND predicates.
                    or
                ListOfOrAndFilters where Filters are expressed in disjunctive normal form (DNF), like
                [[MetadataFilter("short_haired", "==", True), FieldFilter("label", "in", ["cat", "dog"]), ...].
                DNF allows arbitrary boolean logical combinations of single field predicates. The innermost structures
                each describe a single column predicate. The list of inner predicates is interpreted as a conjunction
                (AND), forming a more selective `and` multiple field predicate.
                Finally, the most outer list combines these filters as a disjunction (OR).
            prediction_filters: Filter predicates. Allowed formats are:
                ListOfAndFilters where each Filter forms a chain of AND predicates.
                    or
                ListOfOrAndFilters where Filters are expressed in disjunctive normal form (DNF), like
                [[MetadataFilter("short_haired", "==", True), FieldFilter("label", "in", ["cat", "dog"]), ...].
                DNF allows arbitrary boolean logical combinations of single field predicates. The innermost structures
                each describe a single column predicate. The list of inner predicates is interpreted as a conjunction
                (AND), forming a more selective `and` multiple field predicate.
                Finally, the most outer list combines these filters as a disjunction (OR).
        """
        assert (
            0 <= iou_threshold <= 1
        ), "IoU threshold must be between 0 and 1."
        self.iou_threshold = iou_threshold
        super().__init__(
            enforce_label_match,
            confidence_threshold,
            annotation_filters,
            prediction_filters,
        )

    def eval(
        self,
        annotations: List[BoxOrPolygonAnnotation],
        predictions: List[BoxOrPolygonPrediction],
    ) -> ScalarResult:
        true_positives = num_true_positives(
            annotations, predictions, self.iou_threshold
        )
        weight = len(predictions)
        return ScalarResult(
            true_positives / max(weight, sys.float_info.epsilon), weight
        )


class PolygonRecall(PolygonMetric):
    """Calculates the recall between box or polygon annotations and predictions.
    ::

        from nucleus import BoxAnnotation, Point, PolygonPrediction
        from nucleus.annotation import AnnotationList
        from nucleus.prediction import PredictionList
        from nucleus.metrics import PolygonRecall

        box_anno = BoxAnnotation(
            label="car",
            x=0,
            y=0,
            width=10,
            height=10,
            reference_id="image_1",
            annotation_id="image_1_car_box_1",
            metadata={"vehicle_color": "red"}
        )

        polygon_pred = PolygonPrediction(
            label="bus",
            vertices=[Point(100, 100), Point(150, 200), Point(200, 100)],
            reference_id="image_2",
            annotation_id="image_2_bus_polygon_1",
            confidence=0.8,
            metadata={"vehicle_color": "yellow"}
        )

        annotations = AnnotationList(box_annotations=[box_anno])
        predictions = PredictionList(polygon_predictions=[polygon_pred])
        metric = PolygonRecall()
        metric(annotations, predictions)
    """

    # TODO: Remove defaults once these are surfaced more cleanly to users.
    def __init__(
        self,
        enforce_label_match: bool = False,
        iou_threshold: float = 0.5,
        confidence_threshold: float = 0.0,
        annotation_filters: Optional[
            Union[ListOfOrAndFilters, ListOfAndFilters]
        ] = None,
        prediction_filters: Optional[
            Union[ListOfOrAndFilters, ListOfAndFilters]
        ] = None,
    ):
        """Initializes PolygonRecall object.

        Args:
            enforce_label_match: whether to enforce that annotation and prediction labels must match. Defaults to False
            iou_threshold: IOU threshold to consider detection as valid. Must be in [0, 1]. Default 0.5
            confidence_threshold: minimum confidence threshold for predictions. Must be in [0, 1]. Default 0.0
            annotation_filters: Filter predicates. Allowed formats are:
                ListOfAndFilters where each Filter forms a chain of AND predicates.
                    or
                ListOfOrAndFilters where Filters are expressed in disjunctive normal form (DNF), like
                [[MetadataFilter("short_haired", "==", True), FieldFilter("label", "in", ["cat", "dog"]), ...].
                DNF allows arbitrary boolean logical combinations of single field predicates. The innermost structures
                each describe a single column predicate. The list of inner predicates is interpreted as a conjunction
                (AND), forming a more selective `and` multiple field predicate.
                Finally, the most outer list combines these filters as a disjunction (OR).
            prediction_filters: Filter predicates. Allowed formats are:
                ListOfAndFilters where each Filter forms a chain of AND predicates.
                    or
                ListOfOrAndFilters where Filters are expressed in disjunctive normal form (DNF), like
                [[MetadataFilter("short_haired", "==", True), FieldFilter("label", "in", ["cat", "dog"]), ...].
                DNF allows arbitrary boolean logical combinations of single field predicates. The innermost structures
                each describe a single column predicate. The list of inner predicates is interpreted as a conjunction
                (AND), forming a more selective `and` multiple field predicate.
                Finally, the most outer list combines these filters as a disjunction (OR).
        """
        assert (
            0 <= iou_threshold <= 1
        ), "IoU threshold must be between 0 and 1."
        self.iou_threshold = iou_threshold
        super().__init__(
            enforce_label_match,
            confidence_threshold,
            annotation_filters=annotation_filters,
            prediction_filters=prediction_filters,
        )

    def eval(
        self,
        annotations: List[BoxOrPolygonAnnotation],
        predictions: List[BoxOrPolygonPrediction],
    ) -> ScalarResult:
        true_positives = num_true_positives(
            annotations, predictions, self.iou_threshold
        )
        weight = len(annotations) + sys.float_info.epsilon
        return ScalarResult(
            true_positives / max(weight, sys.float_info.epsilon), weight
        )


class PolygonAveragePrecision(PolygonMetric):
    """Calculates the average precision between box or polygon annotations and predictions.
    ::

        from nucleus import BoxAnnotation, Point, PolygonPrediction
        from nucleus.annotation import AnnotationList
        from nucleus.prediction import PredictionList
        from nucleus.metrics import PolygonAveragePrecision

        box_anno = BoxAnnotation(
            label="car",
            x=0,
            y=0,
            width=10,
            height=10,
            reference_id="image_1",
            annotation_id="image_1_car_box_1",
            metadata={"vehicle_color": "red"}
        )

        polygon_pred = PolygonPrediction(
            label="bus",
            vertices=[Point(100, 100), Point(150, 200), Point(200, 100)],
            reference_id="image_2",
            annotation_id="image_2_bus_polygon_1",
            confidence=0.8,
            metadata={"vehicle_color": "yellow"}
        )

        annotations = AnnotationList(box_annotations=[box_anno])
        predictions = PredictionList(polygon_predictions=[polygon_pred])
        metric = PolygonAveragePrecision(label="car")
        metric(annotations, predictions)
    """

    # TODO: Remove defaults once these are surfaced more cleanly to users.
    def __init__(
        self,
        label,
        iou_threshold: float = 0.5,
        annotation_filters: Optional[
            Union[ListOfOrAndFilters, ListOfAndFilters]
        ] = None,
        prediction_filters: Optional[
            Union[ListOfOrAndFilters, ListOfAndFilters]
        ] = None,
    ):
        """Initializes PolygonRecall object.

        Args:
            iou_threshold: IOU threshold to consider detection as valid. Must be in [0, 1]. Default 0.5
            annotation_filters: Filter predicates. Allowed formats are:
                ListOfAndFilters where each Filter forms a chain of AND predicates.
                    or
                ListOfOrAndFilters where Filters are expressed in disjunctive normal form (DNF), like
                [[MetadataFilter("short_haired", "==", True), FieldFilter("label", "in", ["cat", "dog"]), ...].
                DNF allows arbitrary boolean logical combinations of single field predicates. The innermost structures
                each describe a single column predicate. The list of inner predicates is interpreted as a conjunction
                (AND), forming a more selective `and` multiple field predicate.
                Finally, the most outer list combines these filters as a disjunction (OR).
            prediction_filters: Filter predicates. Allowed formats are:
                ListOfAndFilters where each Filter forms a chain of AND predicates.
                    or
                ListOfOrAndFilters where Filters are expressed in disjunctive normal form (DNF), like
                [[MetadataFilter("short_haired", "==", True), FieldFilter("label", "in", ["cat", "dog"]), ...].
                DNF allows arbitrary boolean logical combinations of single field predicates. The innermost structures
                each describe a single column predicate. The list of inner predicates is interpreted as a conjunction
                (AND), forming a more selective `and` multiple field predicate.
                Finally, the most outer list combines these filters as a disjunction (OR).
        """
        assert (
            0 <= iou_threshold <= 1
        ), "IoU threshold must be between 0 and 1."
        self.iou_threshold = iou_threshold
        self.label = label
        super().__init__(
            enforce_label_match=False,
            confidence_threshold=0,
            annotation_filters=annotation_filters,
            prediction_filters=prediction_filters,
        )

    def eval(
        self,
        annotations: List[BoxOrPolygonAnnotation],
        predictions: List[BoxOrPolygonPrediction],
    ) -> ScalarResult:
        annotations_filtered = polygon_label_filter(annotations, self.label)
        predictions_filtered = polygon_label_filter(predictions, self.label)
        (
            true_false_positives,
            confidences,
        ) = get_true_false_positives_confidences(
            annotations_filtered, predictions_filtered, self.iou_threshold
        )
        idxes = np.argsort(-confidences)
        true_false_positives_sorted = true_false_positives[idxes]
        cumulative_true_positives = np.cumsum(true_false_positives_sorted)
        total_predictions = np.arange(1, len(true_false_positives) + 1)
        precisions = cumulative_true_positives / total_predictions
        recalls = cumulative_true_positives / len(annotations)
        average_precision = compute_average_precision(precisions, recalls)
        weight = 1
        return ScalarResult(average_precision, weight)


class PolygonMAP(PolygonMetric):
    """Calculates the mean average precision between box or polygon annotations and predictions.
    ::

        from nucleus import BoxAnnotation, Point, PolygonPrediction
        from nucleus.annotation import AnnotationList
        from nucleus.prediction import PredictionList
        from nucleus.metrics import PolygonMAP

        box_anno = BoxAnnotation(
            label="car",
            x=0,
            y=0,
            width=10,
            height=10,
            reference_id="image_1",
            annotation_id="image_1_car_box_1",
            metadata={"vehicle_color": "red"}
        )

        polygon_pred = PolygonPrediction(
            label="bus",
            vertices=[Point(100, 100), Point(150, 200), Point(200, 100)],
            reference_id="image_2",
            annotation_id="image_2_bus_polygon_1",
            confidence=0.8,
            metadata={"vehicle_color": "yellow"}
        )

        annotations = AnnotationList(box_annotations=[box_anno])
        predictions = PredictionList(polygon_predictions=[polygon_pred])
        metric = PolygonMAP()
        metric(annotations, predictions)
    """

    # TODO: Remove defaults once these are surfaced more cleanly to users.
    def __init__(
        self,
        iou_threshold: float = 0.5,
        annotation_filters: Optional[
            Union[ListOfOrAndFilters, ListOfAndFilters]
        ] = None,
        prediction_filters: Optional[
            Union[ListOfOrAndFilters, ListOfAndFilters]
        ] = None,
    ):
        """Initializes PolygonRecall object.

        Args:
            iou_threshold: IOU threshold to consider detection as valid. Must be in [0, 1]. Default 0.5
            annotation_filters: Filter predicates. Allowed formats are:
                ListOfAndFilters where each Filter forms a chain of AND predicates.
                    or
                ListOfOrAndFilters where Filters are expressed in disjunctive normal form (DNF), like
                [[MetadataFilter("short_haired", "==", True), FieldFilter("label", "in", ["cat", "dog"]), ...].
                DNF allows arbitrary boolean logical combinations of single field predicates. The innermost structures
                each describe a single column predicate. The list of inner predicates is interpreted as a conjunction
                (AND), forming a more selective `and` multiple field predicate.
                Finally, the most outer list combines these filters as a disjunction (OR).
            prediction_filters: Filter predicates. Allowed formats are:
                ListOfAndFilters where each Filter forms a chain of AND predicates.
                    or
                ListOfOrAndFilters where Filters are expressed in disjunctive normal form (DNF), like
                [[MetadataFilter("short_haired", "==", True), FieldFilter("label", "in", ["cat", "dog"]), ...].
                DNF allows arbitrary boolean logical combinations of single field predicates. The innermost structures
                each describe a single column predicate. The list of inner predicates is interpreted as a conjunction
                (AND), forming a more selective `and` multiple field predicate.
                Finally, the most outer list combines these filters as a disjunction (OR).
        """
        assert (
            0 <= iou_threshold <= 1
        ), "IoU threshold must be between 0 and 1."
        self.iou_threshold = iou_threshold
        super().__init__(
            enforce_label_match=False,
            confidence_threshold=0,
            annotation_filters=annotation_filters,
            prediction_filters=prediction_filters,
        )

    def eval(
        self,
        annotations: List[BoxOrPolygonAnnotation],
        predictions: List[BoxOrPolygonPrediction],
    ) -> ScalarResult:
        grouped_inputs = group_boxes_or_polygons_by_label(
            annotations, predictions
        )
        results: List[ScalarResult] = []
        for label, group in grouped_inputs.items():
            annotations_group, predictions_group = group
            metric = PolygonAveragePrecision(label)
            result = metric.eval(annotations_group, predictions_group)
            results.append(result)
        average_result = ScalarResult.aggregate(results)
        return average_result



import time
from dataclasses import dataclass
from typing import Dict, List

import requests

from nucleus.constants import (
    JOB_CREATION_TIME_KEY,
    JOB_ID_KEY,
    JOB_LAST_KNOWN_STATUS_KEY,
    JOB_TYPE_KEY,
    STATUS_KEY,
)
from nucleus.utils import replace_double_slashes

JOB_POLLING_INTERVAL = 5


@dataclass
class AsyncJob:
    """Object used to check the status or errors of a long running asynchronous operation.

    ::

        import nucleus

        client = nucleus.NucleusClient(YOUR_SCALE_API_KEY)
        dataset = client.get_dataset("ds_bwkezj6g5c4g05gqp1eg")

        # When kicking off an asynchronous job, store the return value as a variable
        job = dataset.append(items=YOUR_DATASET_ITEMS, asynchronous=True)

        # Poll for status or errors
        print(job.status())
        print(job.errors())

        # Block until job finishes
        job.sleep_until_complete()
    """

    job_id: str
    job_last_known_status: str
    job_type: str
    job_creation_time: str
    client: "NucleusClient"  # type: ignore # noqa: F821

    def status(self) -> Dict[str, str]:
        """Fetches status of the job and an informative message on job progress.

        Returns:
            A dict of the job ID, status (one of Running, Completed, or Errored),
            an informative message on the job progress, and number of both completed
            and total steps.
            ::

                {
                    "job_id": "job_c19xcf9mkws46gah0000",
                    "status": "Completed",
                    "message": "Job completed successfully.",
                    "job_progress": "0.33",
                    "completed_steps": "1",
                    "total_steps:": "3",
                }
        """
        response = self.client.make_request(
            payload={},
            route=f"job/{self.job_id}",
            requests_command=requests.get,
        )
        self.job_last_known_status = response[STATUS_KEY]
        return response

    def errors(self) -> List[str]:
        """Fetches a list of the latest errors generated by the asynchronous job.

        Useful for debugging failed or partially successful jobs.

        Returns:
            A list of strings containing the 10,000 most recently generated errors.
            ::

                [
                    '{"annotation":{"label":"car","type":"box","geometry":{"x":50,"y":60,"width":70,"height":80},"referenceId":"bad_ref_id","annotationId":"attempted_annot_upload","metadata":{}},"error":"Item with id bad_ref_id doesn\'t exist."}'
                ]
        """
        errors = self.client.make_request(
            payload={},
            route=f"job/{self.job_id}/errors",
            requests_command=requests.get,
        )
        return [replace_double_slashes(error) for error in errors]

    def sleep_until_complete(self, verbose_std_out=True):
        """Blocks until the job completes or errors.

        Parameters:
            verbose_std_out (Optional[bool]): Whether or not to verbosely log while
              sleeping. Defaults to True.
        """
        start_time = time.perf_counter()
        while 1:
            status = self.status()
            time.sleep(JOB_POLLING_INTERVAL)

            if verbose_std_out:
                print(
                    f"Status at {time.perf_counter() - start_time} s: {status}"
                )
            if status["status"] == "Running":
                continue

            break

        if verbose_std_out:
            print(
                f"Finished at {time.perf_counter() - start_time} s: {status}"
            )
        final_status = status
        if final_status["status"] == "Errored":
            raise JobError(final_status, self)

    @classmethod
    def from_json(cls, payload: dict, client):
        # TODO: make private
        return cls(
            job_id=payload[JOB_ID_KEY],
            job_last_known_status=payload[JOB_LAST_KNOWN_STATUS_KEY],
            job_type=payload[JOB_TYPE_KEY],
            job_creation_time=payload[JOB_CREATION_TIME_KEY],
            client=client,
        )


class JobError(Exception):
    def __init__(self, job_status: Dict[str, str], job: AsyncJob):
        final_status_message = job_status["message"]
        final_status = job_status["status"]
        message = (
            f"The job reported a final status of {final_status} "
            "This could, however, mean a partial success with some successes and some failures. "
            f"The final status message was: {final_status_message} \n"
            f"For more detailed error messages you can call {str(job)}.errors()"
        )
        message = replace_double_slashes(message)
        super().__init__(message)



import click
from rich.live import Live
from rich.spinner import Spinner
from rich.table import Column, Table

from cli.client import init_client
from cli.helpers.web_helper import launch_web_or_invoke


@click.group("jobs", invoke_without_command=True)
@click.option("--web", is_flag=True, help="Launch browser")
@click.pass_context
def jobs(ctx, web):
    """Jobs are a wrapper around various long-running tasks withing Nucleus

    https://dashboard.scale.com/nucleus/jobs
    """
    launch_web_or_invoke("jobs", ctx, web, list_jobs)


@jobs.command("list")
def list_jobs():
    """List all of your Jobs"""
    client = init_client()
    table = Table(
        Column("id", overflow="fold", min_width=24),
        "status",
        "type",
        "created at",
        title=":satellite: Jobs",
        title_justify="left",
    )
    with Live(Spinner("dots4", text="Finding your Jobs!")) as live:
        all_jobs = client.jobs
        for job in all_jobs:
            table.add_row(
                job.job_id,
                job.job_last_known_status,
                job.job_type,
                job.job_creation_time,
            )
            live.update(table)



import itertools
from typing import Callable, Dict, List, Optional, Union

from nucleus.logger import logger
from nucleus.validate.eval_functions.base_eval_function import (
    EvalFunctionConfig,
)

from ...metrics.filtering import ListOfAndFilters, ListOfOrAndFilters
from ..data_transfer_objects.eval_function import EvalFunctionEntry
from ..errors import EvalFunctionNotAvailableError


class PolygonIOUConfig(EvalFunctionConfig):
    def __call__(
        self,
        enforce_label_match: bool = False,
        iou_threshold: float = 0.0,
        confidence_threshold: float = 0.0,
        annotation_filters: Optional[
            Union[ListOfOrAndFilters, ListOfAndFilters]
        ] = None,
        prediction_filters: Optional[
            Union[ListOfOrAndFilters, ListOfAndFilters]
        ] = None,
        **kwargs,
    ):
        """Configures a call to :class:`PolygonIOU` object.
        ::

            import nucleus

            client = nucleus.NucleusClient(YOUR_SCALE_API_KEY)
            bbox_iou: BoundingBoxIOU = client.validate.eval_functions.bbox_iou
            slice_id = "slc_<your_slice>"
            scenario_test = client.validate.create_scenario_test(
                "Example test",
                slice_id=slice_id,
                evaluation_criteria=[bbox_iou(confidence_threshold=0.8) > 0.5]
            )

        Args:
            enforce_label_match: whether to enforce that annotation and prediction labels must match. Defaults to False
            iou_threshold: IOU threshold to consider detection as valid. Must be in [0, 1]. Default 0.0
            confidence_threshold: minimum confidence threshold for predictions. Must be in [0, 1]. Default 0.0
            annotation_filters: Filter predicates. Allowed formats are:
                ListOfAndFilters where each Filter forms a chain of AND predicates.
                    or
                ListOfOrAndFilters where Filters are expressed in disjunctive normal form (DNF), like
                [[MetadataFilter("short_haired", "==", True), FieldFilter("label", "in", ["cat", "dog"]), ...].
                DNF allows arbitrary boolean logical combinations of single field predicates. The innermost structures
                each describe a single column predicate. The list of inner predicates is interpreted as a conjunction
                (AND), forming a more selective `and` multiple field predicate.
                Finally, the most outer list combines these filters as a disjunction (OR).
            prediction_filters: Filter predicates. Allowed formats are:
                ListOfAndFilters where each Filter forms a chain of AND predicates.
                    or
                ListOfOrAndFilters where Filters are expressed in disjunctive normal form (DNF), like
                [[MetadataFilter("short_haired", "==", True), FieldFilter("label", "in", ["cat", "dog"]), ...].
                DNF allows arbitrary boolean logical combinations of single field predicates. The innermost structures
                each describe a single column predicate. The list of inner predicates is interpreted as a conjunction
                (AND), forming a more selective `and` multiple field predicate.
                Finally, the most outer list combines these filters as a disjunction (OR).
        """
        return super().__call__(
            enforce_label_match=enforce_label_match,
            iou_threshold=iou_threshold,
            confidence_threshold=confidence_threshold,
            **kwargs,
        )

    @classmethod
    def expected_name(cls) -> str:
        return "bbox_iou"


class PolygonMAPConfig(EvalFunctionConfig):
    def __call__(
        self,
        iou_threshold: float = 0.5,
        annotation_filters: Optional[
            Union[ListOfOrAndFilters, ListOfAndFilters]
        ] = None,
        prediction_filters: Optional[
            Union[ListOfOrAndFilters, ListOfAndFilters]
        ] = None,
        **kwargs,
    ):
        """Configures a call to :class:`PolygonMAP` object.
        ::

            import nucleus

            client = nucleus.NucleusClient(YOUR_SCALE_API_KEY)
            bbox_map: BoundingBoxMeanAveragePrecision= client.validate.eval_functions.bbox_map
            slice_id = "slc_<your_slice>"
            scenario_test = client.validate.create_scenario_test(
                "Example test",
                slice_id=slice_id,
                evaluation_criteria=[bbox_map(iou_threshold=0.6) > 0.8]
            )

        Args:
            iou_threshold: IOU threshold to consider detection as valid. Must be in [0, 1]. Default 0.0
            annotation_filters: Filter predicates. Allowed formats are:
                ListOfAndFilters where each Filter forms a chain of AND predicates.
                    or
                ListOfOrAndFilters where Filters are expressed in disjunctive normal form (DNF), like
                [[MetadataFilter("short_haired", "==", True), FieldFilter("label", "in", ["cat", "dog"]), ...].
                DNF allows arbitrary boolean logical combinations of single field predicates. The innermost structures
                each describe a single column predicate. The list of inner predicates is interpreted as a conjunction
                (AND), forming a more selective `and` multiple field predicate.
                Finally, the most outer list combines these filters as a disjunction (OR).
            prediction_filters: Filter predicates. Allowed formats are:
                ListOfAndFilters where each Filter forms a chain of AND predicates.
                    or
                ListOfOrAndFilters where Filters are expressed in disjunctive normal form (DNF), like
                [[MetadataFilter("short_haired", "==", True), FieldFilter("label", "in", ["cat", "dog"]), ...].
                DNF allows arbitrary boolean logical combinations of single field predicates. The innermost structures
                each describe a single column predicate. The list of inner predicates is interpreted as a conjunction
                (AND), forming a more selective `and` multiple field predicate.
                Finally, the most outer list combines these filters as a disjunction (OR).
        """
        return super().__call__(
            iou_threshold=iou_threshold,
            **kwargs,
        )

    @classmethod
    def expected_name(cls) -> str:
        return "bbox_map"


class PolygonRecallConfig(EvalFunctionConfig):
    def __call__(
        self,
        enforce_label_match: bool = False,
        iou_threshold: float = 0.5,
        confidence_threshold: float = 0.0,
        annotation_filters: Optional[
            Union[ListOfOrAndFilters, ListOfAndFilters]
        ] = None,
        prediction_filters: Optional[
            Union[ListOfOrAndFilters, ListOfAndFilters]
        ] = None,
        **kwargs,
    ):
        """Configures a call to :class:`PolygonRecall` object.
        ::

            import nucleus

            client = nucleus.NucleusClient(YOUR_SCALE_API_KEY)
            bbox_recall: BoundingBoxMeanAveragePrecision= client.validate.eval_functions.bbox_recall
            slice_id = "slc_<your_slice>"
            scenario_test = client.validate.create_scenario_test(
                "Example test",
                slice_id=slice_id,
                evaluation_criteria=[bbox_recall(iou_threshold=0.6, confidence_threshold=0.4) > 0.9]
            )

        Args:
            enforce_label_match: whether to enforce that annotation and prediction labels must match. Defaults to False
            iou_threshold: IOU threshold to consider detection as valid. Must be in [0, 1]. Default 0.0
            confidence_threshold: minimum confidence threshold for predictions. Must be in [0, 1]. Default 0.0
            annotation_filters: Filter predicates. Allowed formats are:
                ListOfAndFilters where each Filter forms a chain of AND predicates.
                    or
                ListOfOrAndFilters where Filters are expressed in disjunctive normal form (DNF), like
                [[MetadataFilter("short_haired", "==", True), FieldFilter("label", "in", ["cat", "dog"]), ...].
                DNF allows arbitrary boolean logical combinations of single field predicates. The innermost structures
                each describe a single column predicate. The list of inner predicates is interpreted as a conjunction
                (AND), forming a more selective `and` multiple field predicate.
                Finally, the most outer list combines these filters as a disjunction (OR).
            prediction_filters: Filter predicates. Allowed formats are:
                ListOfAndFilters where each Filter forms a chain of AND predicates.
                    or
                ListOfOrAndFilters where Filters are expressed in disjunctive normal form (DNF), like
                [[MetadataFilter("short_haired", "==", True), FieldFilter("label", "in", ["cat", "dog"]), ...].
                DNF allows arbitrary boolean logical combinations of single field predicates. The innermost structures
                each describe a single column predicate. The list of inner predicates is interpreted as a conjunction
                (AND), forming a more selective `and` multiple field predicate.
                Finally, the most outer list combines these filters as a disjunction (OR).
        """
        return super().__call__(
            enforce_label_match=enforce_label_match,
            iou_threshold=iou_threshold,
            confidence_threshold=confidence_threshold,
            **kwargs,
        )

    @classmethod
    def expected_name(cls) -> str:
        return "bbox_recall"


class PolygonPrecisionConfig(EvalFunctionConfig):
    def __call__(
        self,
        enforce_label_match: bool = False,
        iou_threshold: float = 0.5,
        confidence_threshold: float = 0.0,
        annotation_filters: Optional[
            Union[ListOfOrAndFilters, ListOfAndFilters]
        ] = None,
        prediction_filters: Optional[
            Union[ListOfOrAndFilters, ListOfAndFilters]
        ] = None,
        **kwargs,
    ):
        """Configures a call to :class:`PolygonPrecision` object.
        ::

            import nucleus

            client = nucleus.NucleusClient(YOUR_SCALE_API_KEY)
            bbox_precision: BoundingBoxMeanAveragePrecision= client.validate.eval_functions.bbox_precision
            slice_id = "slc_<your_slice>"
            scenario_test = client.validate.create_scenario_test(
                "Example test",
                slice_id=slice_id,
                evaluation_criteria=[bbox_precision(iou_threshold=0.6, confidence_threshold=0.4) > 0.9]
            )

        Args:
            enforce_label_match: whether to enforce that annotation and prediction labels must match. Defaults to False
            iou_threshold: IOU threshold to consider detection as valid. Must be in [0, 1]. Default 0.0
            confidence_threshold: minimum confidence threshold for predictions. Must be in [0, 1]. Default 0.0
            annotation_filters: Filter predicates. Allowed formats are:
                ListOfAndFilters where each Filter forms a chain of AND predicates.
                    or
                ListOfOrAndFilters where Filters are expressed in disjunctive normal form (DNF), like
                [[MetadataFilter("short_haired", "==", True), FieldFilter("label", "in", ["cat", "dog"]), ...].
                DNF allows arbitrary boolean logical combinations of single field predicates. The innermost structures
                each describe a single column predicate. The list of inner predicates is interpreted as a conjunction
                (AND), forming a more selective `and` multiple field predicate.
                Finally, the most outer list combines these filters as a disjunction (OR).
            prediction_filters: Filter predicates. Allowed formats are:
                ListOfAndFilters where each Filter forms a chain of AND predicates.
                    or
                ListOfOrAndFilters where Filters are expressed in disjunctive normal form (DNF), like
                [[MetadataFilter("short_haired", "==", True), FieldFilter("label", "in", ["cat", "dog"]), ...].
                DNF allows arbitrary boolean logical combinations of single field predicates. The innermost structures
                each describe a single column predicate. The list of inner predicates is interpreted as a conjunction
                (AND), forming a more selective `and` multiple field predicate.
                Finally, the most outer list combines these filters as a disjunction (OR).
        """
        return super().__call__(
            enforce_label_match=enforce_label_match,
            iou_threshold=iou_threshold,
            confidence_threshold=confidence_threshold,
            **kwargs,
        )

    @classmethod
    def expected_name(cls) -> str:
        return "bbox_precision"


class CuboidIOU2DConfig(EvalFunctionConfig):
    def __call__(
        self,
        enforce_label_match: bool = True,
        iou_threshold: float = 0.0,
        confidence_threshold: float = 0.0,
        annotation_filters: Optional[
            Union[ListOfOrAndFilters, ListOfAndFilters]
        ] = None,
        prediction_filters: Optional[
            Union[ListOfOrAndFilters, ListOfAndFilters]
        ] = None,
        **kwargs,
    ):
        """Configure a call to CuboidIOU object.

        Args:
            enforce_label_match: whether to enforce that annotation and prediction labels must match. Defaults to True
            iou_threshold: IOU threshold to consider detection as valid. Must be in [0, 1]. Default 0.0
            confidence_threshold: minimum confidence threshold for predictions. Must be in [0, 1]. Default 0.0
            annotation_filters: Filter predicates. Allowed formats are:
                ListOfAndFilters where each Filter forms a chain of AND predicates.
                    or
                ListOfOrAndFilters where Filters are expressed in disjunctive normal form (DNF), like
                [[MetadataFilter("short_haired", "==", True), FieldFilter("label", "in", ["cat", "dog"]), ...].
                DNF allows arbitrary boolean logical combinations of single field predicates. The innermost structures
                each describe a single column predicate. The list of inner predicates is interpreted as a conjunction
                (AND), forming a more selective `and` multiple field predicate.
                Finally, the most outer list combines these filters as a disjunction (OR).
            prediction_filters: Filter predicates. Allowed formats are:
                ListOfAndFilters where each Filter forms a chain of AND predicates.
                    or
                ListOfOrAndFilters where Filters are expressed in disjunctive normal form (DNF), like
                [[MetadataFilter("short_haired", "==", True), FieldFilter("label", "in", ["cat", "dog"]), ...].
                DNF allows arbitrary boolean logical combinations of single field predicates. The innermost structures
                each describe a single column predicate. The list of inner predicates is interpreted as a conjunction
                (AND), forming a more selective `and` multiple field predicate.
                Finally, the most outer list combines these filters as a disjunction (OR).
        """
        return super().__call__(
            enforce_label_match=enforce_label_match,
            iou_threshold=iou_threshold,
            confidence_threshold=confidence_threshold,
            iou_2d=True,
            annotation_filters=annotation_filters,
            prediction_filters=prediction_filters,
            **kwargs,
        )

    @classmethod
    def expected_name(cls) -> str:
        return "cuboid_iou_2d"


class CuboidIOU3DConfig(EvalFunctionConfig):
    def __call__(
        self,
        enforce_label_match: bool = True,
        iou_threshold: float = 0.0,
        confidence_threshold: float = 0.0,
        annotation_filters: Optional[
            Union[ListOfOrAndFilters, ListOfAndFilters]
        ] = None,
        prediction_filters: Optional[
            Union[ListOfOrAndFilters, ListOfAndFilters]
        ] = None,
        **kwargs,
    ):
        """Configure a call to CuboidIOU object.

        Args:
            enforce_label_match: whether to enforce that annotation and prediction labels must match. Defaults to True
            iou_threshold: IOU threshold to consider detection as valid. Must be in [0, 1]. Default 0.0
            iou_2d: whether to return the BEV 2D IOU if true, or the 3D IOU if false.
            confidence_threshold: minimum confidence threshold for predictions. Must be in [0, 1]. Default 0.0
            annotation_filters: Filter predicates. Allowed formats are:
                ListOfAndFilters where each Filter forms a chain of AND predicates.
                    or
                ListOfOrAndFilters where Filters are expressed in disjunctive normal form (DNF), like
                [[MetadataFilter("short_haired", "==", True), FieldFilter("label", "in", ["cat", "dog"]), ...].
                DNF allows arbitrary boolean logical combinations of single field predicates. The innermost structures
                each describe a single column predicate. The list of inner predicates is interpreted as a conjunction
                (AND), forming a more selective `and` multiple field predicate.
                Finally, the most outer list combines these filters as a disjunction (OR).
            prediction_filters: Filter predicates. Allowed formats are:
                ListOfAndFilters where each Filter forms a chain of AND predicates.
                    or
                ListOfOrAndFilters where Filters are expressed in disjunctive normal form (DNF), like
                [[MetadataFilter("short_haired", "==", True), FieldFilter("label", "in", ["cat", "dog"]), ...].
                DNF allows arbitrary boolean logical combinations of single field predicates. The innermost structures
                each describe a single column predicate. The list of inner predicates is interpreted as a conjunction
                (AND), forming a more selective `and` multiple field predicate.
                Finally, the most outer list combines these filters as a disjunction (OR).
        """
        return super().__call__(
            enforce_label_match=enforce_label_match,
            iou_threshold=iou_threshold,
            confidence_threshold=confidence_threshold,
            iou_2d=False,
            annotation_filters=annotation_filters,
            prediction_filters=prediction_filters,
            **kwargs,
        )

    @classmethod
    def expected_name(cls) -> str:
        return "cuboid_iou_3d"


class CuboidPrecisionConfig(EvalFunctionConfig):
    def __call__(
        self,
        enforce_label_match: bool = True,
        iou_threshold: float = 0.0,
        confidence_threshold: float = 0.0,
        annotation_filters: Optional[
            Union[ListOfOrAndFilters, ListOfAndFilters]
        ] = None,
        prediction_filters: Optional[
            Union[ListOfOrAndFilters, ListOfAndFilters]
        ] = None,
        **kwargs,
    ):
        """Configure a call to CuboidPrecision object.

        Args:
            enforce_label_match: whether to enforce that annotation and prediction labels must match. Defaults to True
            iou_threshold: IOU threshold to consider detection as valid. Must be in [0, 1]. Default 0.0
            iou_2d: whether to return the BEV 2D IOU if true, or the 3D IOU if false.
            confidence_threshold: minimum confidence threshold for predictions. Must be in [0, 1]. Default 0.0
            annotation_filters: Filter predicates. Allowed formats are:
                ListOfAndFilters where each Filter forms a chain of AND predicates.
                    or
                ListOfOrAndFilters where Filters are expressed in disjunctive normal form (DNF), like
                [[MetadataFilter("short_haired", "==", True), FieldFilter("label", "in", ["cat", "dog"]), ...].
                DNF allows arbitrary boolean logical combinations of single field predicates. The innermost structures
                each describe a single column predicate. The list of inner predicates is interpreted as a conjunction
                (AND), forming a more selective `and` multiple field predicate.
                Finally, the most outer list combines these filters as a disjunction (OR).
            prediction_filters: Filter predicates. Allowed formats are:
                ListOfAndFilters where each Filter forms a chain of AND predicates.
                    or
                ListOfOrAndFilters where Filters are expressed in disjunctive normal form (DNF), like
                [[MetadataFilter("short_haired", "==", True), FieldFilter("label", "in", ["cat", "dog"]), ...].
                DNF allows arbitrary boolean logical combinations of single field predicates. The innermost structures
                each describe a single column predicate. The list of inner predicates is interpreted as a conjunction
                (AND), forming a more selective `and` multiple field predicate.
                Finally, the most outer list combines these filters as a disjunction (OR).
        """
        return super().__call__(
            enforce_label_match=enforce_label_match,
            iou_threshold=iou_threshold,
            confidence_threshold=confidence_threshold,
            annotation_filters=annotation_filters,
            prediction_filters=prediction_filters,
            **kwargs,
        )

    @classmethod
    def expected_name(cls) -> str:
        return "cuboid_precision"


class CuboidRecallConfig(EvalFunctionConfig):
    def __call__(
        self,
        enforce_label_match: bool = True,
        iou_threshold: float = 0.0,
        confidence_threshold: float = 0.0,
        annotation_filters: Optional[
            Union[ListOfOrAndFilters, ListOfAndFilters]
        ] = None,
        prediction_filters: Optional[
            Union[ListOfOrAndFilters, ListOfAndFilters]
        ] = None,
        **kwargs,
    ):
        """Configure a call to a CuboidRecall object.

        Args:
            enforce_label_match: whether to enforce that annotation and prediction labels must match. Defaults to True
            iou_threshold: IOU threshold to consider detection as valid. Must be in [0, 1]. Default 0.0
            iou_2d: whether to return the BEV 2D IOU if true, or the 3D IOU if false.
            confidence_threshold: minimum confidence threshold for predictions. Must be in [0, 1]. Default 0.0
            annotation_filters: Filter predicates. Allowed formats are:
                ListOfAndFilters where each Filter forms a chain of AND predicates.
                    or
                ListOfOrAndFilters where Filters are expressed in disjunctive normal form (DNF), like
                [[MetadataFilter("short_haired", "==", True), FieldFilter("label", "in", ["cat", "dog"]), ...].
                DNF allows arbitrary boolean logical combinations of single field predicates. The innermost structures
                each describe a single column predicate. The list of inner predicates is interpreted as a conjunction
                (AND), forming a more selective `and` multiple field predicate.
                Finally, the most outer list combines these filters as a disjunction (OR).
            prediction_filters: Filter predicates. Allowed formats are:
                ListOfAndFilters where each Filter forms a chain of AND predicates.
                    or
                ListOfOrAndFilters where Filters are expressed in disjunctive normal form (DNF), like
                [[MetadataFilter("short_haired", "==", True), FieldFilter("label", "in", ["cat", "dog"]), ...].
                DNF allows arbitrary boolean logical combinations of single field predicates. The innermost structures
                each describe a single column predicate. The list of inner predicates is interpreted as a conjunction
                (AND), forming a more selective `and` multiple field predicate.
                Finally, the most outer list combines these filters as a disjunction (OR).
        """
        return super().__call__(
            enforce_label_match=enforce_label_match,
            iou_threshold=iou_threshold,
            confidence_threshold=confidence_threshold,
            annotation_filters=annotation_filters,
            prediction_filters=prediction_filters,
            **kwargs,
        )

    @classmethod
    def expected_name(cls) -> str:
        return "cuboid_recall"


class CategorizationF1Config(EvalFunctionConfig):
    def __call__(
        self,
        confidence_threshold: Optional[float] = None,
        f1_method: Optional[str] = None,
        annotation_filters: Optional[
            Union[ListOfOrAndFilters, ListOfAndFilters]
        ] = None,
        prediction_filters: Optional[
            Union[ListOfOrAndFilters, ListOfAndFilters]
        ] = None,
        **kwargs,
    ):
        """ Configure an evaluation of :class:`CategorizationF1`.
        ::

            import nucleus

            client = nucleus.NucleusClient(YOUR_SCALE_API_KEY)
            cat_f1: CategorizationF1 = client.validate.eval_functions.cat_f1
            slice_id = "slc_<your_slice>"
            scenario_test = client.validate.create_scenario_test(
                "Example test",
                slice_id=slice_id,
                evaluation_criteria=[cat_f1(confidence_threshold=0.6, f1_method="weighted") > 0.7]
            )

        Args:
            confidence_threshold: minimum confidence threshold for predictions to be taken into account for evaluation.
                 Must be in [0, 1]. Default 0.0
            f1_method: {'micro', 'macro', 'samples','weighted', 'binary'}, \
                default='macro'
            This parameter is required for multiclass/multilabel targets.
            If ``None``, the scores for each class are returned. Otherwise, this
            determines the type of averaging performed on the data:

            ``'binary'``:
                Only report results for the class specified by ``pos_label``.
                This is applicable only if targets (``y_{true,pred}``) are binary.
            ``'micro'``:
                Calculate metrics globally by counting the total true positives,
                false negatives and false positives.
            ``'macro'``:
                Calculate metrics for each label, and find their unweighted
                mean.  This does not take label imbalance into account.
            ``'weighted'``:
                Calculate metrics for each label, and find their average weighted
                by support (the number of true instances for each label). This
                alters 'macro' to account for label imbalance; it can result in an
                F-score that is not between precision and recall.
            ``'samples'``:
                Calculate metrics for each instance, and find their average (only
                meaningful for multilabel classification where this differs from
                :func:`accuracy_score`).
            annotation_filters: Filter predicates. Allowed formats are:
                ListOfAndFilters where each Filter forms a chain of AND predicates.
                    or
                ListOfOrAndFilters where Filters are expressed in disjunctive normal form (DNF), like
                [[MetadataFilter("short_haired", "==", True), FieldFilter("label", "in", ["cat", "dog"]), ...].
                DNF allows arbitrary boolean logical combinations of single field predicates. The innermost structures
                each describe a single column predicate. The list of inner predicates is interpreted as a conjunction
                (AND), forming a more selective `and` multiple field predicate.
                Finally, the most outer list combines these filters as a disjunction (OR).
            prediction_filters: Filter predicates. Allowed formats are:
                ListOfAndFilters where each Filter forms a chain of AND predicates.
                    or
                ListOfOrAndFilters where Filters are expressed in disjunctive normal form (DNF), like
                [[MetadataFilter("short_haired", "==", True), FieldFilter("label", "in", ["cat", "dog"]), ...].
                DNF allows arbitrary boolean logical combinations of single field predicates. The innermost structures
                each describe a single column predicate. The list of inner predicates is interpreted as a conjunction
                (AND), forming a more selective `and` multiple field predicate.
                Finally, the most outer list combines these filters as a disjunction (OR).
        """
        return super().__call__(
            confidence_threshold=confidence_threshold, f1_method=f1_method
        )

    @classmethod
    def expected_name(cls) -> str:
        return "cat_f1"


class CustomEvalFunction(EvalFunctionConfig):
    @classmethod
    def expected_name(cls) -> str:
        raise NotImplementedError(
            "Custm evaluation functions are coming soon"
        )  # Placeholder: See super().eval_func_entry for actual name


class StandardEvalFunction(EvalFunctionConfig):
    """Class for standard Model CI eval functions that have not been added as attributes on
    AvailableEvalFunctions yet.
    """

    def __init__(self, eval_function_entry: EvalFunctionEntry):
        logger.warning(
            "Standard function %s not implemented as an attribute on AvailableEvalFunctions",
            eval_function_entry.name,
        )
        super().__init__(eval_function_entry)

    @classmethod
    def expected_name(cls) -> str:
        return "public_function"  # Placeholder: See super().eval_func_entry for actual name


class EvalFunctionNotAvailable(EvalFunctionConfig):
    def __init__(
        self, not_available_name: str
    ):  # pylint: disable=super-init-not-called
        self.not_available_name = not_available_name

    def __call__(self, *args, **kwargs):
        self._raise_error()

    def _op_to_test_metric(self, *args, **kwargs):
        self._raise_error()

    def _raise_error(self):
        raise EvalFunctionNotAvailableError(
            f"Eval function '{self.not_available_name}' is not available to the current user. "
            f"Is Model CI enabled for the user?"
        )

    @classmethod
    def expected_name(cls) -> str:
        return "public_function"  # Placeholder: See super().eval_func_entry for actual name


EvalFunction = Union[
    PolygonIOUConfig,
    PolygonMAPConfig,
    PolygonPrecisionConfig,
    PolygonRecallConfig,
    CuboidRecallConfig,
    CuboidIOU2DConfig,
    CuboidIOU3DConfig,
    CuboidPrecisionConfig,
    CategorizationF1Config,
    CustomEvalFunction,
    EvalFunctionNotAvailable,
    StandardEvalFunction,
]


class AvailableEvalFunctions:
    """Collection class that acts as a common entrypoint to access evaluation functions. Standard evaluation functions
    provided by Scale are attributes of this class.

    The available evaluation functions are listed in the sample below::

        e = client.validate.eval_functions
        unit_test_criteria = [
            e.bbox_iou() > 5,
            e.bbox_map() > 0.95,
            e.bbox_precision() > 0.8,
            e.bbox_recall() > 0.5,
        ]
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(self, available_functions: List[EvalFunctionEntry]):
        assert (
            available_functions
        ), "Passed no available functions for current user. Is the feature flag enabled?"
        self._public_func_entries: Dict[str, EvalFunctionEntry] = {
            f.name: f for f in available_functions if f.is_public
        }
        # NOTE: Public are assigned
        self._public_to_function: Dict[str, EvalFunctionConfig] = {}
        self._custom_to_function: Dict[str, CustomEvalFunction] = {
            f.name: CustomEvalFunction(f)
            for f in available_functions
            if not f.is_public
        }
        self.bbox_iou: PolygonIOUConfig = self._assign_eval_function_if_defined(PolygonIOUConfig)  # type: ignore
        self.bbox_precision: PolygonPrecisionConfig = self._assign_eval_function_if_defined(
            PolygonPrecisionConfig  # type: ignore
        )
        self.bbox_recall: PolygonRecallConfig = self._assign_eval_function_if_defined(
            PolygonRecallConfig  # type: ignore
        )
        self.bbox_map: PolygonMAPConfig = self._assign_eval_function_if_defined(
            PolygonMAPConfig  # type: ignore
        )
        self.cat_f1: CategorizationF1Config = self._assign_eval_function_if_defined(
            CategorizationF1Config  # type: ignore
        )
        self.cuboid_iou_2d: CuboidIOU2DConfig = self._assign_eval_function_if_defined(CuboidIOU2DConfig)  # type: ignore
        self.cuboid_iou_3d: CuboidIOU3DConfig = self._assign_eval_function_if_defined(CuboidIOU3DConfig)  # type: ignore
        self.cuboid_precision: CuboidPrecisionConfig = self._assign_eval_function_if_defined(CuboidPrecisionConfig)  # type: ignore
        self.cuboid_recall: CuboidRecallConfig = self._assign_eval_function_if_defined(CuboidRecallConfig)  # type: ignore

        # Add public entries that have not been implemented as an attribute on this class
        for func_entry in self._public_func_entries.values():
            if func_entry.name not in self._public_to_function:
                self._public_to_function[
                    func_entry.name
                ] = StandardEvalFunction(func_entry)

    def __repr__(self):
        """Standard functions are ones Scale provides and custom ones customer defined"""
        # NOTE: setting to lower to be consistent with attribute names
        functions_lower = [
            str(name).lower() for name in self._public_func_entries.keys()
        ]
        return (
            f"<AvailableEvaluationFunctions: public:{functions_lower} "
            f"private: {list(self._custom_to_function.keys())}"
        )

    @property
    def public_functions(self) -> Dict[str, EvalFunctionConfig]:
        """Standard functions provided by Model CI.

        Notes:
            These functions are also available as attributes on :class:`AvailableEvalFunctions`

        Returns:
            Dict of function name to :class:`BaseEvalFunction`.
        """
        return self._public_to_function

    @property
    def private_functions(self) -> Dict[str, CustomEvalFunction]:
        """Custom functions uploaded to Model CI

        Returns:
            Dict of function name to :class:`CustomEvalFunction`.
        """
        return self._custom_to_function

    def _assign_eval_function_if_defined(
        self,
        eval_function_constructor: Callable[[EvalFunctionEntry], EvalFunction],
    ):
        """Helper function for book-keeping and assignment of standard Scale provided functions that are accessible
        via attribute access
        """
        # TODO(gunnar): Too convoluted .. simplify
        expected_name = eval_function_constructor.expected_name()  # type: ignore
        if expected_name in self._public_func_entries:
            definition = self._public_func_entries[expected_name]
            eval_function = eval_function_constructor(definition)
            self._public_to_function[expected_name] = eval_function  # type: ignore
            return eval_function
        else:
            return EvalFunctionNotAvailable(expected_name)

    def from_id(self, eval_function_id: str):
        for eval_func in itertools.chain(
            self._public_to_function.values(),
            self._custom_to_function.values(),
        ):
            if eval_func.id == eval_function_id:
                return eval_func
        raise EvalFunctionNotAvailableError(
            f"Could not find Eval Function with id {eval_function_id}"
        )



from typing import Any, Dict, List, Optional

from nucleus.pydantic_base import DictCompatibleModel


class DatasetInfo(DictCompatibleModel):
    """High-level :class:`Dataset` information

    Attributes:
        dataset_id: Nucleus-generated dataset ID
        name: User-defined name of dataset
        length: Number of :class:`DatasetItem` in :class:`Dataset`
        model_run_ids: (deprecated)
        slice_ids: List :class:`Slice` IDs associated with the :class:`Dataset`
        annotation_metadata_schema: Dict defining annotation-level metadata schema.
        item_metadata_schema: Dict defining item metadata schema.
    """

    dataset_id: str
    name: str
    length: int
    model_run_ids: List[str]
    slice_ids: List[str]
    # TODO: Expand the following into pydantic models to formalize schema
    annotation_metadata_schema: Optional[Dict[str, Any]] = None
    item_metadata_schema: Optional[Dict[str, Any]] = None



import click
from rich.console import Console
from rich.live import Live
from rich.spinner import Spinner
from rich.table import Column, Table
from rich.tree import Tree

from cli.client import init_client
from cli.helpers.nucleus_url import nucleus_url
from cli.helpers.web_helper import launch_web_or_invoke
from nucleus import NucleusAPIError


@click.group("slices", invoke_without_command=True)
@click.option("--web", is_flag=True, help="Launch browser")
@click.pass_context
def slices(ctx, web):
    """Slices are named subsets of Datasets"""
    # TODO(gunnar): We don't have a natural landing for slices overview, until then we land on "/"
    launch_web_or_invoke("", ctx, web, list_slices)


@slices.command("list")
def list_slices():
    """List all available Slices"""
    with Live(
        Spinner("dots4", text="Finding your Slices!"),
        vertical_overflow="visible",
    ) as live:
        client = init_client()
        datasets = client.datasets
        table = Table(
            Column("id", overflow="fold", min_width=24),
            "name",
            "dataset_name",
            Column("url", overflow="fold"),
            title=":cake: Slices",
            title_justify="left",
        )
        errors = {}
        for ds in datasets:
            try:
                ds_slices = ds.slices
                if ds_slices:
                    for slc_id in ds_slices:
                        slice_url = nucleus_url(f"{ds.id}/{slc_id}")
                        slice_info = client.get_slice(slc_id).info()
                        table.add_row(
                            slc_id, slice_info["name"], ds.name, slice_url
                        )
                        live.update(table)
            except NucleusAPIError as e:
                errors[ds.id] = e

        error_tree = Tree(
            ":x: Encountered the following errors while fetching information"
        )
        for ds_id, error in errors.items():
            dataset_branch = error_tree.add(f"Dataset: {ds_id}")
            dataset_branch.add(f"Error: {error}")

        Console().print(error_tree)



from .base import Metric, ScalarResult
from .categorization_metrics import CategorizationF1
from .cuboid_metrics import CuboidIOU, CuboidPrecision, CuboidRecall
from .filtering import (
    FieldFilter,
    ListOfOrAndFilters,
    MetadataFilter,
    apply_filters,
)
from .polygon_metrics import (
    PolygonAveragePrecision,
    PolygonIOU,
    PolygonMAP,
    PolygonMetric,
    PolygonPrecision,
    PolygonRecall,
)



import logging

import requests

logger = logging.getLogger(__name__)
logging.basicConfig()
logging.getLogger(
    requests.packages.urllib3.__package__  # pylint: disable=no-member
).setLevel(logging.ERROR)



"""Compute active learning metrics on your predictions.

For more details on usage see the example colab in scripts/autocurate_bdd.ipynb
"""


import datetime

import requests

from nucleus.constants import (
    JOB_CREATION_TIME_KEY,
    JOB_LAST_KNOWN_STATUS_KEY,
    JOB_TYPE_KEY,
)
from nucleus.job import AsyncJob


def entropy(name, model_run, client):
    """Computes the mean entropy across all predictions for each image."""
    model_run_ids = [model_run.model_run_id]
    dataset_id = model_run.dataset_id
    response = client.make_request(
        payload={"modelRunIds": model_run_ids},
        route=f"autocurate/{dataset_id}/single_model_entropy/{name}",
        requests_command=requests.post,
    )
    # TODO: the response should already have the below three fields populated
    response[JOB_LAST_KNOWN_STATUS_KEY] = "Started"
    response[JOB_TYPE_KEY] = "autocurateEntropy"
    response[JOB_CREATION_TIME_KEY] = (
        datetime.datetime.now().isoformat("T", "milliseconds") + "Z"
    )
    job = AsyncJob.from_json(response, client)
    return job



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cli',
 'cli.helpers',
 'nucleus',
 'nucleus.data_transfer_object',
 'nucleus.metrics',
 'nucleus.validate',
 'nucleus.validate.data_transfer_objects',
 'nucleus.validate.eval_functions']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.7.4,<4.0.0',
 'click>=7.1.2,<9.0',
 'nest-asyncio>=1.5.1,<2.0.0',
 'numpy>=1.19.5,<2.0.0',
 'pydantic>=1.8.2,<2.0.0',
 'requests>=2.23.0,<3.0.0',
 'rich>=10.15.2,<11.0.0',
 'scikit-learn>=0.24.0',
 'scipy>=1.4.1',
 'shellingham>=1.4.0,<2.0.0',
 'tqdm>=4.41.0,<5.0.0']

extras_require = \
{':python_full_version >= "3.6.1" and python_version < "3.7"': ['dataclasses>=0.7,<0.8'],
 'shapely': ['Shapely>=1.8.0']}

entry_points = \
{'console_scripts': ['nu = cli.nu:nu']}

setup_kwargs = {
    'name': 'scale-nucleus',
    'version': '0.9b6',
    'description': 'The official Python client library for Nucleus, the Data Platform for AI',
    'long_description': '# Nucleus\n\nhttps://dashboard.scale.com/nucleus\n\nAggregate metrics in ML are not good enough. To improve production ML, you need to understand their qualitative failure modes, fix them by gathering more data, and curate diverse scenarios.\n\nScale Nucleus helps you:\n\n- Visualize your data\n- Curate interesting slices within your dataset\n- Review and manage annotations\n- Measure and debug your model performance\n\nNucleus is a new way—the right way—to develop ML models, helping us move away from the concept of one dataset and towards a paradigm of collections of scenarios.\n\n## Installation\n\n`$ pip install scale-nucleus`\n\n\n## CLI installation\nWe recommend installing the CLI via `pipx` (https://pypa.github.io/pipx/installation/). This makes sure that\nthe CLI does not interfere with you system packages and is accessible from your favorite terminal.\n\nFor MacOS:\n```bash\nbrew install pipx\npipx ensurepath\npipx install scale-nucleus\n# Optional installation of shell completion (for bash, zsh or fish)\nnu install-completions\n```\n\nOtherwise, install via pip (requires pip 19.0 or later):\n```bash\npython3 -m pip install --user pipx\npython3 -m pipx ensurepath\npython3 -m pipx install scale-nucleus\n# Optional installation of shell completion (for bash, zsh or fish)\nnu install-completions\n```\n\n## Common issues/FAQ\n\n### Outdated Client\n\nNucleus is iterating rapidly and as a result we do not always perfectly preserve backwards compatibility with older versions of the client. If you run into any unexpected error, it\'s a good idea to upgrade your version of the client by running\n```\npip install --upgrade scale-nucleus\n```\n\n## Usage\n\nFor the most up to date documentation, reference: https://dashboard.scale.com/nucleus/docs/api?language=python.\n\n## For Developers\n\nClone from github and install as editable\n\n```\ngit clone git@github.com:scaleapi/nucleus-python-client.git\ncd nucleus-python-client\npip3 install poetry\npoetry install\n```\n\nPlease install the pre-commit hooks by running the following command:\n\n```python\npoetry run pre-commit install\n```\n\nWhen releasing a new version please add release notes to the changelog in `CHANGELOG.md`.\n\n**Best practices for testing:**\n(1). Please run pytest from the root directory of the repo, i.e.\n\n```\npoetry run pytest tests/test_dataset.py\n```\n\n(2) To skip slow integration tests that have to wait for an async job to start.\n\n```\npoetry run pytest -m "not integration"\n```\n\n## Pydantic Models\n\nPrefer using [Pydantic](https://pydantic-docs.helpmanual.io/usage/models/) models rather than creating raw dictionaries \nor dataclasses to send or receive over the wire as JSONs. Pydantic is created with data validation in mind and provides very clear error \nmessages when it encounters a problem with the payload.\n\nThe Pydantic model(s) should mirror the payload to send. To represent a JSON payload that looks like this:\n```json\n{\n  "example_json_with_info": {\n      "metadata": {\n        "frame": 0\n      },\n      "reference_id": "frame0",\n      "url": "s3://example/scale_nucleus/2021/lidar/0038711321865000.json",\n      "type": "pointcloud"\n    },\n  "example_image_with_info": {\n      "metadata": {\n        "author": "Picasso"\n      },\n      "reference_id": "frame0",\n      "url": "s3://bucket/0038711321865000.jpg",\n      "type": "image"\n    },\n}\n```\n\nCould be represented as the following structure. Note that the field names map to the JSON keys and the usage of field \nvalidators (`@validator`).\n\n```python\nimport os.path\nfrom pydantic import BaseModel, validator\nfrom typing import Literal\n\n\nclass JsonWithInfo(BaseModel):\n    metadata: dict  # any dict is valid\n    reference_id: str\n    url: str\n    type: Literal["pointcloud", "recipe"]\n\n    @validator("url")\n    def has_json_extension(cls, v):\n        if not v.endswith(".json"):\n            raise ValueError(f"Expected \'.json\' extension got {v}")\n        return v\n\n\nclass ImageWithInfo(BaseModel):\n    metadata: dict  # any dict is valid\n    reference_id: str\n    url: str\n    type: Literal["image", "mask"]\n\n    @validator("url")\n    def has_valid_extension(cls, v):\n        valid_extensions = {".jpg", ".jpeg", ".png", ".tiff"}\n        _, extension = os.path.splitext(v)\n        if extension not in valid_extensions:\n            raise ValueError(f"Expected extension in {valid_extensions} got {v}")\n        return v\n\n\nclass ExampleNestedModel(BaseModel):\n    example_json_with_info: JsonWithInfo\n    example_image_with_info: ImageWithInfo\n\n# Usage:\nimport requests\npayload = requests.get("/example")\nparsed_model = ExampleNestedModel.parse_obj(payload.json())\nrequests.post("example/post_to", json=parsed_model.dict())\n```\n\n\n### Migrating to Pydantic\n- When migrating an interface from a dictionary use `nucleus.pydantic_base.DictCompatibleModel`. That allows you to get\nthe benefits of Pydantic but maintaints backwards compatibility with a Python dictionary by delegating `__getitem__` to \nfields.\n- When migrating a frozen dataclass use `nucleus.pydantic_base.ImmutableModel`. That is a base class set up to be \nimmutable after initialization.\n\n**Updating documentation:**\nWe use [Sphinx](https://www.sphinx-doc.org/en/master/) to autogenerate our API Reference from docstrings.\n\nTo test your local docstring changes, run the following commands from the repository\'s root directory:\n```\npoetry shell\ncd docs\nsphinx-autobuild . ./_build/html --watch ../nucleus\n```\n`sphinx-autobuild` will spin up a server on localhost (port 8000 by default) that will watch for and automatically rebuild a version of the API reference based on your local docstring changes.\n\n\n## Custom Metrics used in Scale Validate\n\nTo install the availalbe metrics for Scale Validate add the optional `metrics` extra.\nNote that you might need to install a local GEOS package.\n\n```bash\n#Mac OS\nbrew install geos\n# Ubuntu/Debian flavors\napt-get install libgeos-dev\n```\n\n`pip install scale-nucleus[metrics]`\n\nTo develop it locally use\n\n`poetry install --extra metrics`\n',
    'author': 'Scale AI Nucleus Team',
    'author_email': 'nucleusapi@scaleapi.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://scale.com/nucleus',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)



from typing import Any, Dict, List, Optional, Union

from .annotation import (
    BoxAnnotation,
    CategoryAnnotation,
    CuboidAnnotation,
    MultiCategoryAnnotation,
    PolygonAnnotation,
    SegmentationAnnotation,
)
from .constants import (
    ANNOTATION_METADATA_SCHEMA_KEY,
    ANNOTATION_UPDATE_KEY,
    ANNOTATIONS_KEY,
    ITEMS_KEY,
    LABELS_KEY,
    METADATA_KEY,
    MODEL_ID_KEY,
    NAME_KEY,
    REFERENCE_ID_KEY,
    SCENES_KEY,
    SEGMENTATIONS_KEY,
    TAXONOMY_NAME_KEY,
    TYPE_KEY,
    UPDATE_KEY,
)
from .dataset_item import DatasetItem
from .prediction import (
    BoxPrediction,
    CategoryPrediction,
    CuboidPrediction,
    PolygonPrediction,
    SegmentationPrediction,
)
from .scene import LidarScene, VideoScene


def construct_append_payload(
    dataset_items: List[DatasetItem], force: bool = False
) -> dict:
    items = []
    for item in dataset_items:
        items.append(item.to_payload())

    return (
        {ITEMS_KEY: items}
        if not force
        else {ITEMS_KEY: items, UPDATE_KEY: True}
    )


def construct_append_scenes_payload(
    scene_list: Union[List[LidarScene], List[VideoScene]],
    update: Optional[bool] = False,
) -> dict:
    scenes = []
    for scene in scene_list:
        scenes.append(scene.to_payload())
    return {SCENES_KEY: scenes, UPDATE_KEY: update}


def construct_annotation_payload(
    annotation_items: List[
        Union[
            BoxAnnotation,
            PolygonAnnotation,
            CuboidAnnotation,
            CategoryAnnotation,
            MultiCategoryAnnotation,
            SegmentationAnnotation,
        ]
    ],
    update: bool,
) -> dict:
    annotations = [
        annotation.to_payload()
        for annotation in annotation_items
        if not isinstance(annotation, SegmentationAnnotation)
    ]
    segmentations = [
        annotation.to_payload()
        for annotation in annotation_items
        if isinstance(annotation, SegmentationAnnotation)
    ]
    payload: Dict[str, Any] = {ANNOTATION_UPDATE_KEY: update}
    if annotations:
        payload[ANNOTATIONS_KEY] = annotations
    if segmentations:
        payload[SEGMENTATIONS_KEY] = segmentations
    return payload


def construct_segmentation_payload(
    annotation_items: Union[
        List[SegmentationAnnotation], List[SegmentationPrediction]
    ],
    update: bool,
) -> dict:
    annotations = []
    for annotation_item in annotation_items:
        annotations.append(annotation_item.to_payload())

    return {SEGMENTATIONS_KEY: annotations, ANNOTATION_UPDATE_KEY: update}


def construct_box_predictions_payload(
    box_predictions: List[
        Union[
            BoxPrediction,
            PolygonPrediction,
            CuboidPrediction,
            CategoryPrediction,
        ]
    ],
    update: bool,
) -> dict:
    predictions = []
    for prediction in box_predictions:
        predictions.append(prediction.to_payload())

    return {ANNOTATIONS_KEY: predictions, ANNOTATION_UPDATE_KEY: update}


def construct_model_creation_payload(
    name: str, reference_id: str, metadata: Optional[Dict]
) -> dict:
    return {
        NAME_KEY: name,
        REFERENCE_ID_KEY: reference_id,
        METADATA_KEY: metadata if metadata else {},
    }


def construct_model_run_creation_payload(
    name: str,
    reference_id: Optional[str],
    model_id: Optional[str],
    metadata: Optional[Dict],
    annotation_metadata_schema: Optional[Dict] = None,
) -> dict:
    payload = {
        NAME_KEY: name,
        METADATA_KEY: metadata if metadata else {},
    }
    if reference_id:
        payload[REFERENCE_ID_KEY] = reference_id
    if model_id:
        payload[MODEL_ID_KEY] = model_id
    return {
        NAME_KEY: name,
        REFERENCE_ID_KEY: reference_id,
        METADATA_KEY: metadata if metadata else {},
        ANNOTATION_METADATA_SCHEMA_KEY: annotation_metadata_schema,
    }


def construct_taxonomy_payload(
    taxonomy_name: str, taxonomy_type: str, labels: List[str], update: bool
) -> dict:
    return {
        TAXONOMY_NAME_KEY: taxonomy_name,
        TYPE_KEY: taxonomy_type,
        LABELS_KEY: labels,
        UPDATE_KEY: update,
    }



from typing import List

from nucleus.prediction import PredictionList

from .polygon_utils import (
    BoxOrPolygonAnnoOrPred,
    polygon_annotation_to_geometry,
)


def polygon_area_filter(
    polygons: List[BoxOrPolygonAnnoOrPred], min_area: float, max_area: float
) -> List[BoxOrPolygonAnnoOrPred]:
    filter_fn = (
        lambda polygon: min_area
        <= polygon_annotation_to_geometry(polygon).signed_area
        <= max_area
    )
    return list(filter(filter_fn, polygons))


def confidence_filter(
    predictions: PredictionList, min_confidence: float
) -> PredictionList:
    predictions_copy = PredictionList()
    filter_fn = (
        lambda prediction: not hasattr(prediction, "confidence")
        or prediction.confidence >= min_confidence
    )
    for attr in predictions.__dict__:
        predictions_copy.__dict__[attr] = list(
            filter(filter_fn, predictions.__dict__[attr])
        )
    return predictions_copy


def polygon_label_filter(
    polygons: List[BoxOrPolygonAnnoOrPred], label: str
) -> List[BoxOrPolygonAnnoOrPred]:
    return list(filter(lambda polygon: polygon.label == label, polygons))



from enum import Enum

EVALUATION_ID_KEY = "evaluation_id"
EVAL_FUNCTION_ID_KEY = "eval_function_id"
ID_KEY = "id"
PASS_KEY = "pass"
RESULT_KEY = "result"
THRESHOLD_COMPARISON_KEY = "threshold_comparison"
THRESHOLD_KEY = "threshold"
SCENARIO_TEST_ID_KEY = "scenario_test_id"
SCENARIO_TEST_NAME_KEY = "scenario_test_name"
SCENARIO_TEST_METRICS_KEY = "scenario_test_metrics"


class ThresholdComparison(str, Enum):
    """Comparator between the result and the threshold."""

    GREATER_THAN = "greater_than"
    GREATER_THAN_EQUAL_TO = "greater_than_equal_to"
    LESS_THAN = "less_than"
    LESS_THAN_EQUAL_TO = "less_than_equal_to"



import click

from cli.helpers.nucleus_url import nucleus_url


def launch_web_or_show_help(
    sub_url: str, ctx: click.Context, launch_browser: bool
):
    """Launches the sub_url (composed with nuclues_url(sub_url)) in the browser if requested"""
    if not ctx.invoked_subcommand:
        if launch_browser:
            url = nucleus_url(sub_url)
            click.launch(url)
        else:
            click.echo(ctx.get_help())
    else:
        if launch_browser:
            click.echo(click.style("--web does not work with sub-commands"))
            ctx.abort()


def launch_web_or_invoke(
    sub_url: str,
    ctx: click.Context,
    launch_browser: bool,
    command: click.BaseCommand,
):
    """Launches the sub_url (composed with nuclues_url(sub_url)) in the browser if requested, otherwise invokes
    the passed command
    """
    if not ctx.invoked_subcommand:
        if launch_browser:
            url = nucleus_url(sub_url)
            click.launch(url)
        else:
            ctx.invoke(command)
    else:
        if launch_browser:
            click.echo(click.style("--web does not work with sub-commands"))
            ctx.abort()
